'''
This module holds the customization settings for the debugger.
'''
from _pydevd_bundle.pydevd_constants import QUOTED_LINE_PROTOCOL


class PydevdCustomization(object):
    DEFAULT_PROTOCOL = QUOTED_LINE_PROTOCOL
